ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


def caesar_enc(letter: str, k: int = 3) -> str:
    """Encrypts a single uppercase letter using the Caesar cipher with a
    given shift.

    Args:
        letter (str): A string of length 1 representing an uppercase letter.
        k (int): An integer representing the shift amount.
            Optional; defaults to 3.

    Returns:
        str: A string of length 1 representing the encrypted letter.

    Raises:
        ValueError: If the letter is not a valid uppercase letter.

    Example:
        >>> caesar_enc("A", 3)
        'D'
        >>> caesar_enc("Z", -3)
        'W'
        >>> caesar_enc("a", 3)
        [...]
        ValueError: Invalid letter
    """
    if not "A" <= letter <= "Z":
        raise ValueError("Invalid letter")
    return ALPHABET[(ord(letter) - ord("A") + k) % len(ALPHABET)]


def caesar_enc_string(plaintext: str, k: int = 3) -> str:
    """Encrypts a string of uppercase letters using the Caesar cipher with a
    given shift.

    Args:
        plaintext (str): A string of uppercase letters.
        k (int): An integer representing the shift amount.
            Optional; defaults to 3.

    Returns:
        str: A string of uppercase letters representing the encrypted text.

    Raises:
        ValueError: If the plaintext contains any invalid characters.

    Example:
        >>> caesar_enc_string("HELLO", 3)
        'KHOOR'
        >>> caesar_enc_string("WORLD", -3)
        'TLOIA'
        >>> caesar_enc_string("HELLO WORLD", 3)
        [...]
        ValueError: Invalid letter
    """
    ciphertext = ""
    for letter in plaintext:
        ciphertext += caesar_enc(letter, k)
    return ciphertext


def caesar_dec(letter: str, k: int = 3) -> str:
    """Decrypts a single uppercase letter using the inverse Caesar cipher with
    a given shift.

    Args:
        letter (str): A string of length 1 representing an uppercase letter.
        k (int): An integer representing the shift amount.
            Optional; defaults to 3.

    Returns:
        str: A string of length 1 representing the decrypted letter.

    Raises:
        ValueError: If the letter is not a valid uppercase letter.

    Example:
        >>> caesar_dec("D", 3)
        'A'
        >>> caesar_dec("C", -3)
        'F'
        >>> caesar_dec("d", 3)
        [...]
        ValueError: Invalid letter
    """
    if not "A" <= letter <= "Z":
        raise ValueError("Invalid letter")
    return ALPHABET[(ord(letter) - ord("A") - k) % len(ALPHABET)]


def caesar_dec_string(ciphertext: str, k: int = 3) -> str:
    """Decrypts a string of uppercase letters using the inverse Caesar cipher
    with a given shift.

    Args:
        ciphertext (str): A string of uppercase letters.
        k (int): An integer representing the shift amount.
            Optional; defaults to 3.

    Returns:
        str: A string of uppercase letters representing the decrypted text.

    Raises:
        ValueError: If the ciphertext contains any invalid characters.

    Example:
        >>> caesar_dec_string("KHOOR", 3)
        'HELLO'
        >>> caesar_dec_string("TLOIA", -3)
        'WORLD'
        >>> caesar_dec_string("KHOOR WORLD", 3)
        [...]
        ValueError: Invalid letter
    """
    plaintext = ""
    for letter in ciphertext:
        plaintext += caesar_dec(letter, k)
    return plaintext
